@@ -21,10 +21,18 @@ class APIException(PRAWException):
         :param message: The associated message for the error.
         :param field: The input field associated with the error if available.
 
+        .. note:: Calling ``str()`` on the instance returns
+            ``unicode_escape``-d ASCII string because the message may be
+            localized and may contain UNICODE characters. If you want a
+            non-escaped message, access the ``message`` attribute on
+            the instance.
+
         """
-        error_str = '{}: \'{}\''.format(error_type, message)
+        error_str = u'{}: \'{}\''.format(error_type, message)
         if field:
-            error_str += ' on field \'{}\''.format(field)
+            error_str += u' on field \'{}\''.format(field)
+        error_str = error_str.encode('unicode_escape').decode('ascii')
+
         super(APIException, self).__init__(error_str)
         self.error_type = error_type
         self.message = message
