@@ -1,11 +1,14 @@
 """Provide the Subreddit class."""
+# pylint: disable=too-many-lines
 from copy import deepcopy
-from json import dumps
+from json import dumps, loads
+from os.path import basename, dirname, join
 
 from prawcore import Redirect
+import websocket
 
 from ...const import API_PATH, urljoin
-from ...exceptions import APIException
+from ...exceptions import APIException, ClientException
 from ..util import permissions_string, stream_generator
 from ..listing.generator import ListingGenerator
 from ..listing.mixins import SubredditListingMixin
@@ -407,6 +410,79 @@ class Subreddit(RedditBase, MessageableMixin, SubredditListingMixin):
     def _info_path(self):
         return API_PATH['subreddit_about'].format(subreddit=self)
 
+    def _submit_media(self, data):
+        """Submit and return an `image`, `video`, or `videogif`.
+
+        This is a helper method for submitting posts that are not link posts or
+        self posts.
+        """
+        response = self._reddit.post(API_PATH['submit'], data=data)
+
+        # About the websockets:
+        #
+        # Reddit responds to this request with only two fields: a link to
+        # the user's /submitted page, and a websocket URL. We can use the
+        # websocket URL to get a link to the new post once it is created.
+        #
+        # An important note to PRAW contributors or anyone who would
+        # wish to step through this section with a debugger: This block
+        # of code is NOT debugger-friendly. If there is *any*
+        # significant time between the POST request just above this
+        # comment and the creation of the websocket connection just
+        # below, the code will become stuck in an infinite loop at the
+        # socket.recv() call. I believe this is because only one message is
+        # sent over the websocket, and if the client doesn't connect
+        # soon enough, it will miss the message and get stuck forever
+        # waiting for another.
+        #
+        # So if you need to debug this section of code, please let the
+        # websocket creation happen right after the POST request,
+        # otherwise you will have trouble.
+
+        if not isinstance(response, dict):
+            raise ClientException('Something went wrong with your post: '
+                                  '{!r}'.format(response))
+
+        try:
+            socket = websocket.create_connection(response['json']['data']
+                                                 ['websocket_url'],
+                                                 timeout=2)
+            ws_update = loads(socket.recv())
+            socket.close(timeout=2)
+        except websocket.WebSocketTimeoutException:
+            raise ClientException('Websocket error. Check your media file. '
+                                  'Your post may still have been created.')
+        url = ws_update['payload']['redirect']
+        return self._reddit.submission(url=url)
+
+    def _upload_media(self, media_path):
+        """Upload media and return its URL. Uses undocumented endpoint."""
+        if media_path is None:
+            media_path = join(dirname(dirname(dirname(__file__))),
+                              'images', 'PRAW logo.png')
+
+        file_name = basename(media_path).lower()
+        mime_type = {'png': 'image/png', 'mov': 'video/quicktime',
+                     'mp4': 'video/mp4', 'jpg': 'image/jpeg',
+                     'jpeg': 'image/jpeg',
+                     'gif': 'image/gif'}.get(file_name.rpartition('.')[2],
+                                             'image/jpeg')  # default to JPEG
+        img_data = {'filepath': file_name, 'mimetype': mime_type}
+
+        url = API_PATH['media_asset']
+        # until we learn otherwise, assume this request always succeeds
+        upload_lease = self._reddit.post(url, data=img_data)['args']
+        upload_url = 'https:{}'.format(upload_lease['action'])
+        upload_data = {item['name']: item['value']
+                       for item in upload_lease['fields']}
+
+        with open(media_path, 'rb') as media:
+            response = self._reddit._core._requestor._http.post(
+                upload_url, data=upload_data, files={'file': media})
+        response.raise_for_status()
+
+        return upload_url + '/' + upload_data['key']
+
     def random(self):
         """Return a random Submission."""
         url = API_PATH['subreddit_random'].format(subreddit=self)
@@ -506,6 +582,11 @@ class Subreddit(RedditBase, MessageableMixin, SubredditListingMixin):
            url = 'https://praw.readthedocs.io'
            reddit.subreddit('reddit_api_test').submit(title, url=url)
 
+        .. note ::
+
+           For submitting images, videos, and videogifs,
+           see :meth:`.submit_image` and :meth:`.submit_video`.
+
         """
         if (bool(selftext) or selftext == '') == bool(url):
             raise TypeError('Either `selftext` or `url` must be provided.')
@@ -521,6 +602,99 @@ class Subreddit(RedditBase, MessageableMixin, SubredditListingMixin):
             data.update(kind='link', url=url)
         return self._reddit.post(API_PATH['submit'], data=data)
 
+    def submit_image(self, title, image_path, flair_id=None,
+                     flair_text=None, resubmit=True, send_replies=True):
+        """Add an image submission to the subreddit.
+
+        :param title: The title of the submission.
+        :param image_path: The path to an image, to upload and post.
+        :param flair_id: The flair template to select (default: None).
+        :param flair_text: If the template's ``flair_text_editable`` value is
+            True, this value will set a custom text (default: None).
+        :param resubmit: When False, an error will occur if the URL has already
+            been submitted (default: True).
+        :param send_replies: When True, messages will be sent to the submission
+            author when comments are made to the submission (default: True).
+
+        .. note::
+
+           Reddit's API uses WebSockets to respond with the link of the
+           newly created post. If this fails, the method will raise
+           :class:`.ClientException`. Occasionally, the Reddit post will still
+           be created. More often, there is an error with the image file.
+
+        :returns: A :class:`~.Submission` object for the newly created
+            submission.
+
+        For example to submit an image to ``/r/reddit_api_test`` do:
+
+        .. code:: python
+
+           title = 'My favorite picture'
+           image = '/path/to/image.png'
+           reddit.subreddit('reddit_api_test').submit_image(title, image)
+
+        """
+        data = {'sr': str(self), 'resubmit': bool(resubmit),
+                'sendreplies': bool(send_replies), 'title': title}
+        for key, value in (('flair_id', flair_id), ('flair_text', flair_text)):
+            if value is not None:
+                data[key] = value
+        data.update(kind='image', url=self._upload_media(image_path))
+        return self._submit_media(data)
+
+    def submit_video(self, title, video_path, videogif=False,
+                     thumbnail_path=None, flair_id=None, flair_text=None,
+                     resubmit=True, send_replies=True):
+        """Add a video or videogif submission to the subreddit.
+
+        :param title: The title of the submission.
+        :param video_path: The path to a video, to upload and post.
+        :param videogif: A ``bool`` value. If ``True``, the video is
+            uploaded as a videogif, which is essentially a silent video (
+            default: ``False``).
+        :param thumbnail_path: (Optional) The path to an image, to be uploaded
+            and used as the thumbnail for this video. If not provided, the
+            PRAW logo will be used as the thumbnail.
+        :param flair_id: The flair template to select (default: ``None``).
+        :param flair_text: If the template's ``flair_text_editable`` value is
+            True, this value will set a custom text (default: ``None``).
+        :param resubmit: When False, an error will occur if the URL has already
+            been submitted (default: ``True``).
+        :param send_replies: When True, messages will be sent to the submission
+            author when comments are made to the submission
+            (default: ``True``).
+
+        .. note::
+
+           Reddit's API uses WebSockets to respond with the link of the
+           newly created post. If this fails, the method will raise
+           :class:`.ClientException`. Occasionally, the Reddit post will still
+           be created. More often, there is an error with the video file.
+
+        :returns: A :class:`~.Submission` object for the newly created
+            submission.
+
+        For example to submit a video to ``/r/reddit_api_test`` do:
+
+        .. code:: python
+
+           title = 'My favorite movie'
+           video = '/path/to/video.mp4'
+           reddit.subreddit('reddit_api_test').submit_video(title, video)
+
+        """
+        data = {'sr': str(self), 'resubmit': bool(resubmit),
+                'sendreplies': bool(send_replies), 'title': title}
+        for key, value in (('flair_id', flair_id), ('flair_text', flair_text)):
+            if value is not None:
+                data[key] = value
+        data.update(kind='videogif' if videogif else 'video',
+                    url=self._upload_media(video_path),
+                    # if thumbnail_path is None, it uploads the PRAW logo
+                    video_poster_url=self._upload_media(thumbnail_path))
+        return self._submit_media(data)
+
     def subscribe(self, other_subreddits=None):
         """Subscribe to the subreddit.
 
@@ -843,10 +1017,25 @@ class SubredditFlairTemplates(object):
         """
         self.subreddit = subreddit
 
-    def _add(self, text, css_class='', text_editable=False, is_link=None):
-        url = API_PATH['flairtemplate'].format(subreddit=self.subreddit)
-        data = {'css_class': css_class, 'flair_type': self.flair_type(is_link),
-                'text': text, 'text_editable': bool(text_editable)}
+    def _add(self, text, css_class='', text_editable=False, is_link=None,
+             background_color=None, text_color=None, mod_only=None):
+        if css_class and any(param is not None for param in
+                             (background_color, text_color, mod_only)):
+            raise TypeError('Parameter `css_class` cannot be used in '
+                            'conjunction with parameters `background_color`, '
+                            '`text_color`, or `mod_only`.')
+        if css_class:
+            url = API_PATH['flairtemplate'].format(subreddit=self.subreddit)
+            data = {'css_class': css_class,
+                    'flair_type': self.flair_type(is_link),
+                    'text': text, 'text_editable': bool(text_editable)}
+        else:
+            url = API_PATH['flairtemplate_v2'].format(subreddit=self.subreddit)
+            data = {'background_color': background_color,
+                    'text_color': text_color,
+                    'flair_type': self.flair_type(is_link),
+                    'text': text, 'text_editable': bool(text_editable),
+                    'mod_only': bool(mod_only)}
         self.subreddit._reddit.post(url, data=data)
 
     def _clear(self, is_link=None):
@@ -862,35 +1051,64 @@ class SubredditFlairTemplates(object):
         .. code-block:: python
 
            template_info = list(subreddit.flair.templates)[0]
-           subreddit.flair.templates.delete(template_info['flair_template_id])
+           subreddit.flair.templates.delete(template_info['id'])
 
         """
         url = API_PATH['flairtemplatedelete'].format(subreddit=self.subreddit)
         self.subreddit._reddit.post(
             url, data={'flair_template_id': template_id})
 
-    def update(self, template_id, text, css_class='', text_editable=False):
-        """Update the flair templated provided by ``template_id``.
+    def update(self, template_id, text, css_class='', text_editable=False,
+               background_color=None, text_color=None, mod_only=None):
+        """Update the flair template provided by ``template_id``.
 
         :param template_id: The flair template to update.
         :param text: The flair template's new text (required).
         :param css_class: The flair template's new css_class (default: '').
+            Cannot be used in conjunction with ``background_color``,
+            ``text_color``, or ``mod_only``.
         :param text_editable: (boolean) Indicate if the flair text can be
             modified for each Redditor that sets it (default: False).
+        :param background_color: The flair template's new background color,
+            as a hex color. Cannot be used in conjunction with ``css_class``.
+        :param text_color: The flair template's new text color, either
+            ``'light'`` or ``'dark'``. Cannot be used in conjunction with
+            ``css_class``.
+        :param mod_only: (boolean) Indicate if the flair can only be used by
+            moderators. Cannot be used in conjunction with ``css_class``.
 
-        For example to make a link flair template text_editable, try:
+        For example to make a user flair template text_editable, try:
 
         .. code-block:: python
 
            template_info = list(subreddit.flair.templates)[0]
            subreddit.flair.templates.update(
-               template_info['flair_template_id'],
+               template_info['id'],
+               template_info['flair_text'],
                text_editable=True)
 
+        .. note::
+
+           Any parameters not provided will be set to default values (usually
+           ``None`` or ``False``) on Reddit's end.
+
         """
-        url = API_PATH['flairtemplate'].format(subreddit=self.subreddit)
-        data = {'css_class': css_class, 'flair_template_id': template_id,
-                'text': text, 'text_editable': bool(text_editable)}
+        if css_class and any(param is not None for param in
+                             (background_color, text_color, mod_only)):
+            raise TypeError('Parameter `css_class` cannot be used in '
+                            'conjunction with parameters `background_color`, '
+                            '`text_color`, or `mod_only`.')
+
+        if css_class:
+            url = API_PATH['flairtemplate'].format(subreddit=self.subreddit)
+            data = {'css_class': css_class, 'flair_template_id': template_id,
+                    'text': text, 'text_editable': bool(text_editable)}
+        else:
+            url = API_PATH['flairtemplate_v2'].format(subreddit=self.subreddit)
+            data = {'flair_template_id': template_id, 'text': text,
+                    'background_color': background_color,
+                    'text_color': text_color, 'text_editable': text_editable,
+                    'mod_only': mod_only}
         self.subreddit._reddit.post(url, data=data)
 
 
@@ -909,18 +1127,28 @@ class SubredditRedditorFlairTemplates(SubredditFlairTemplates):
 
 
         """
-        url = API_PATH['flairselector'].format(subreddit=self.subreddit)
-        data = {'unique': self.subreddit._reddit._next_unique}
-        for template in self.subreddit._reddit.post(url, data=data)['choices']:
+        url = API_PATH['user_flair'].format(subreddit=self.subreddit)
+        params = {'unique': self.subreddit._reddit._next_unique}
+        for template in self.subreddit._reddit.get(url, params=params):
             yield template
 
-    def add(self, text, css_class='', text_editable=False):
+    def add(self, text, css_class='', text_editable=False,
+            background_color=None, text_color=None, mod_only=None):
         """Add a Redditor flair template to the associated subreddit.
 
         :param text: The flair template's text (required).
         :param css_class: The flair template's css_class (default: '').
+            Cannot be used in conjunction with ``background_color``,
+            ``text_color``, or ``mod_only``.
         :param text_editable: (boolean) Indicate if the flair text can be
             modified for each Redditor that sets it (default: False).
+        :param background_color: The flair template's new background color,
+            as a hex color. Cannot be used in conjunction with ``css_class``.
+        :param text_color: The flair template's new text color, either
+            ``'light'`` or ``'dark'``. Cannot be used in conjunction with
+            ``css_class``.
+        :param mod_only: (boolean) Indicate if the flair can only be used by
+            moderators. Cannot be used in conjunction with ``css_class``.
 
         For example, to add an editable Redditor flair try:
 
@@ -931,7 +1159,8 @@ class SubredditRedditorFlairTemplates(SubredditFlairTemplates):
 
         """
         self._add(text, css_class=css_class, text_editable=text_editable,
-                  is_link=False)
+                  is_link=False, background_color=background_color,
+                  text_color=text_color, mod_only=mod_only)
 
     def clear(self):
         """Remove all Redditor flair templates from the subreddit.
@@ -965,13 +1194,23 @@ class SubredditLinkFlairTemplates(SubredditFlairTemplates):
         for template in self.subreddit._reddit.get(url):
             yield template
 
-    def add(self, text, css_class='', text_editable=False):
+    def add(self, text, css_class='', text_editable=False,
+            background_color=None, text_color=None, mod_only=None):
         """Add a link flair template to the associated subreddit.
 
         :param text: The flair template's text (required).
         :param css_class: The flair template's css_class (default: '').
+            Cannot be used in conjunction with ``background_color``,
+            ``text_color``, or ``mod_only``.
         :param text_editable: (boolean) Indicate if the flair text can be
             modified for each Redditor that sets it (default: False).
+        :param background_color: The flair template's new background color,
+            as a hex color. Cannot be used in conjunction with ``css_class``.
+        :param text_color: The flair template's new text color, either
+            ``'light'`` or ``'dark'``. Cannot be used in conjunction with
+            ``css_class``.
+        :param mod_only: (boolean) Indicate if the flair can only be used by
+            moderators. Cannot be used in conjunction with ``css_class``.
 
         For example, to add an editable link flair try:
 
@@ -982,7 +1221,8 @@ class SubredditLinkFlairTemplates(SubredditFlairTemplates):
 
         """
         self._add(text, css_class=css_class, text_editable=text_editable,
-                  is_link=True)
+                  is_link=True, background_color=background_color,
+                  text_color=text_color, mod_only=mod_only)
 
     def clear(self):
         """Remove all link flair templates from the subreddit.
@@ -1542,8 +1782,7 @@ class ModeratorRelationship(SubredditRelationship):
            reddit.subreddit('subredditname').moderator.leave()
 
         """
-        self.subreddit._reddit.post(API_PATH['leavemoderator'],
-                                    data={'id': self.subreddit.fullname})
+        self.remove(self.subreddit._reddit.config.username)
 
     def remove_invite(self, redditor):
         """Remove the moderator invite for ``redditor``.
@@ -1841,6 +2080,15 @@ class SubredditStream(object):
            for comment in reddit.subreddit('iama').stream.comments():
                print(comment)
 
+        To only retreive new submissions starting when the stream is
+        created, pass `skip_existing=True`:
+
+        .. code:: python
+
+           subreddit = reddit.subreddit('iama')
+           for comment in subreddit.stream.comments(skip_existing=True):
+               print(comment)
+
         """
         return stream_generator(self.subreddit.comments, **stream_options)
 
